# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kink', 'kink.errors']

package_data = \
{'': ['*']}

install_requires = \
['typing_extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'kink',
    'version': '0.3.3',
    'description': 'Dependency injection for python.',
    'long_description': '# Kink [![Build Status](https://travis-ci.org/kodemore/kink.svg?branch=master)](https://travis-ci.org/kodemore/kink) [![codecov](https://codecov.io/gh/kodemore/kink/branch/master/graph/badge.svg)](https://codecov.io/gh/kodemore/kink)\nDependency injection made for python\n\n## Features\n\n- Easy to use interface\n- Extensible with custom dependency resolvers\n- Automatic dependency injection\n- Lightweight\n\n\n## Installation\n\n```\npip install kink\n```\n\n# Usage\n\n## Adding service to dependency injection container\n\nDependency container is a dict-like object, adding new service to dependency container is as \nsimple as the following example:\n\n```python\nfrom kink import di\nfrom os import getenv\n\ndi["db_name"] = getenv("DB_NAME")\ndi["db_password"] = getenv("DB_PASSWORD")\n```\n\n## Adding service factory to dependency injection container\n\nKink also supports on-demand service creation. In order to define such a service, lambda function\nshould be used: \n\n```python\nfrom kink import di\nfrom sqlite3 import connect\n\ndi["db_connection"] = lambda di: connect(di["db_name"])\n```\n\nIn this scenario connection to database will not be established until service is requested.\n\n## Requesting services fromm dependency injection container\n\n```python\nfrom kink import di\nfrom sqlite3 import connect\n\n# Setting services\ndi["db_name"] = "test_db.db"\ndi["db_connection"] = lambda di: connect(di["db_name"])\n\n\n# Getting service\n\nconnection = di["db_connection"] # will return instance of sqlite3.Connection\nassert connection == di["db_connection"] # True\n```\n\n## Autowiring dependencies\n\n```python\nfrom kink import di, inject\nfrom sqlite3 import connect, Connection\n\n\ndi["db_name"] = "test_db.db"\ndi["db_connection"] = lambda di: connect(di["db_name"])\n\n# Inject connection from di, connection is established once function is called.\n@inject\ndef get_database(db_connection: Connection):\n    ...\n\n\nconnection = get_database()\nconnection_with_passed_connection = get_database(connect("temp.db")) # will use passed connection\n```\n\n### Constructor injection\n```python\nfrom kink import inject, di\nimport MySQLdb\n\n# Set dependencies\ndi["db_host"] = "localhost"\ndi["db_name"] = "test"\ndi["db_user"] = "user"\ndi["db_password"] = "password"\ndi["db_connection"] = lambda di: MySQLdb.connect(host=di["db_host"], user=di["db_user"], passwd=di["db_password"], db=di["db_name"])\n\n@inject\nclass AbstractRepository:\n    def __init__(self, db_connection):\n        self.connection = db_connection\n\n\nclass UserRepository(AbstractRepository):\n    ...\n\n\nrepository = UserRepository()\nrepository.connection # mysql db connection is resolved and available to use.\n```\n\nWhen class is annotated by `inject` annotation it will be automatically added to the container for future use (eg autowiring).\n\n\n### Services aliasing\n\nWhen you registering service with `@inject` decorator you can attach your own alias name, please consider the following example:\n\n```python\nfrom kink import inject\nfrom typing import Protocol\n\nclass IUserRepository(Protocol):\n    ...\n\n@inject(alias=IUserRepository)\nclass UserRepository:\n    ...\n\n\nassert di[IUserRepository] == di[UserRepository] # returns true\n```\n\nFor more examples check [tests](/tests) directory\n',
    'author': 'Dawid Kraczkowski',
    'author_email': 'dawid.kraczkowski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kodemore/kink',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
import asyncio
from inspect import isclass, signature
from typing import Any, Callable, Dict, Tuple, Type, TypeVar

from .container import di
from .errors import ExecutionError

T = TypeVar("T")


def _inspect_function_arguments(
    function: Callable,
) -> Tuple[Tuple[str, ...], Dict[str, type]]:
    argument_names: Tuple[str, ...] = tuple(signature(function).parameters.keys())
    argument_types = {}

    for name in argument_names:
        if name in function.__annotations__:
            argument_types[name] = function.__annotations__[name]
        else:
            argument_types[name] = Any

    return argument_names, argument_types


def _resolve_function_kwargs(
    alias_map: Dict[str, str],
    argument_names: Tuple[str, ...],
    argument_types: Dict[str, type],
) -> Dict[str, Any]:
    resolved_kwargs = {}
    for name in argument_names:
        if name in alias_map and alias_map[name] in di:
            resolved_kwargs[name] = di[alias_map[name]]
            continue

        if name in di:
            resolved_kwargs[name] = di[name]
            continue

        if argument_types[name] in di:
            resolved_kwargs[name] = di[argument_types[name]]
            continue

    return resolved_kwargs


def _decorate(binding: Dict[str, Any], service: Type[T]) -> Type[T]:  # type: ignore

    # ignore abstract class initializer
    if service == ABC.__init__:
        return service

    # Add class definition to dependency injection
    argument_names, argument_types = _inspect_function_arguments(service)
    cached_kwargs: Dict[str, Any] = {}

    def _resolve_kwargs(args, kwargs) -> dict:
        nonlocal cached_kwargs

        if not cached_kwargs:
            cached_kwargs = _resolve_function_kwargs(
                binding, argument_names, argument_types
            )

        # attach named arguments
        passed_kwargs = {**kwargs}

        # resolve positional arguments
        if args:
            for key, value in enumerate(args):
                passed_kwargs[argument_names[key]] = value

        all_kwargs = {**cached_kwargs, **passed_kwargs}

        if len(all_kwargs) < len(argument_names):
            missing_parameters = [
                arg for arg in argument_names if arg not in all_kwargs
            ]
            raise ExecutionError(
                "Cannot execute function without required parameters. "
                + f"Did you forget to bind the following parameters: `{'`, `'.join(missing_parameters)}`?"
            )

        return all_kwargs

    def _decorated(*args, **kwargs):
        # all arguments were passed
        if len(args) == len(argument_names):
            return service(*args)

        all_kwargs = _resolve_kwargs(args, kwargs)
        return service(**all_kwargs)

    async def _async_decorated(*args, **kwargs):
        # all arguments were passed
        if len(args) == len(argument_names):
            return await service(*args)

        all_kwargs = _resolve_kwargs(args, kwargs)
        return await service(**all_kwargs)

    if asyncio.iscoroutinefunction(service):
        return _async_decorated  # type: ignore

    return _decorated  # type: ignore


def inject(
    _service: Type[T] = None, alias: Any = None, bind: Dict[str, Any] = None
) -> Type[T]:
    def _decorator(_service: Type[T]) -> Type[T]:
        if isclass(_service):
            di[_service] = lambda _di: _service()
            setattr(
                _service,
                "__init__",
                _decorate(bind or {}, getattr(_service, "__init__")),
            )
            if alias:
                di[alias] = lambda _di: _di[_service]

            return _service
        service_function = _decorate(bind or {}, _service)
        if alias:
            di[alias] = service_function

        return service_function

    if _service is None:
        return _decorator  # type: ignore

    return _decorator(_service)


__all__ = ["inject"]
